import responses
from autoextract.constants import API_ENDPOINT
from autoextract.batching import build_query
from autoextract.sync import request_raw


@responses.activate
def test_sync_request_success():
    responses.add(responses.POST, API_ENDPOINT, json={'success': 'true'})
    query = build_query(urls=['http://example.com'], page_type='article')

    resp = request_raw(query, 'secret_key')
    assert resp == {'success': 'true'}


@responses.activate
def test_sync_request_error():
    responses.add(responses.POST, API_ENDPOINT, json={'error': 'true'}, status=429)
    query = build_query(urls=['http://example.com'], page_type='article')

    try:
        request_raw(query, 'secret_key')
    except Exception as err:
        assert err.response.status_code == 429
        assert 'User-Agent' in err.request.headers



# -*- coding: utf-8 -*-
import copy

import pytest

from autoextract.batching import build_query, record_order

URLS = [
    'http://example.com',
    'https://foo.example.com/foo',
    'http://example.com/1'
]


def test_build_query():
    query = build_query(
        urls=['http://example.com', 'http://example.org'],
        page_type='article')
    assert query == [
        {'url': 'http://example.com', 'pageType': 'article'},
        {'url': 'http://example.org', 'pageType': 'article'},
    ]


def test_build_query_too_many():
    urls = URLS * 100
    with pytest.raises(ValueError):
        build_query(urls, "article")


def test_record_order():
    query = build_query(URLS, "article")
    query_old = copy.deepcopy(query)
    new_query = record_order(query)
    assert query == query_old
    assert [q['meta'] for q in new_query] == ['0', '1', '2']
    assert [q['url'] for q in new_query] == URLS
    assert [q['pageType'] for q in new_query] == ['article'] * len(URLS)



# -*- coding: utf-8 -*-
import os
from typing import Optional

from .constants import ENV_VARIABLE


class NoApiKey(Exception):
    pass


def get_apikey(key: Optional[str] = None) -> str:
    """ Return API key, probably loading it from an environment variable """
    if key is not None:
        return key
    try:
        return os.environ[ENV_VARIABLE]
    except KeyError:
        raise NoApiKey("API key not found. Please set {} "
                       "environment variable or pass".format(ENV_VARIABLE))



# -*- coding: utf-8 -*-
import pytest
import os

from autoextract.apikey import ENV_VARIABLE


@pytest.fixture()
def autoextract_env_variable():
    _FAKE_KEY = 'APIKEY'
    old_env = os.environ.copy()
    os.environ[ENV_VARIABLE] = _FAKE_KEY
    try:
        yield _FAKE_KEY
    finally:
        os.environ = old_env



# -*- coding: utf-8 -*-
"""
Synchronous Scrapinghub AutoExtract API client.
"""

from typing import Optional, Dict, Any, List

import requests

from .batching import record_order, build_query, restore_order
from .constants import API_ENDPOINT, API_TIMEOUT
from .apikey import get_apikey
from .utils import user_agent


def request_raw(query: List[Dict[str, Any]],
                api_key: Optional[str] = None,
                endpoint: str = API_ENDPOINT,
                ) -> List[Dict[str, Any]]:
    """ Send a request to Scrapinghub AutoExtract API.
    Query is a list of dicts, as described in the API docs
    (see https://doc.scrapinghub.com/autoextract.html).
    """
    auth = (get_apikey(api_key), '')
    timeout = API_TIMEOUT + 60
    headers = {'User-Agent': user_agent(requests)}
    resp = requests.post(endpoint, json=query, auth=auth,
                         headers=headers, timeout=timeout)
    resp.raise_for_status()
    return resp.json()


def request_batch(urls: List[str],
                  page_type: str,
                  api_key: Optional[str] = None,
                  endpoint: str = API_ENDPOINT,
                  ) -> List[Dict]:
    query = record_order(build_query(urls, page_type))
    results = request_raw(query, api_key=api_key, endpoint=endpoint)
    return restore_order(results)



# -*- coding: utf-8 -*-
""" Helpers for batching requests """
from copy import deepcopy
from typing import List, Dict

from .constants import API_MAX_BATCH


def build_query(urls: List[str], page_type: str) -> List[Dict]:
    """ Given a list of URLs and page type, return query """
    if len(urls) > API_MAX_BATCH:
        raise ValueError("Batch size can't be greater than %s" %
                         API_MAX_BATCH)
    return [{'url': url, 'pageType': page_type} for url in urls]


def record_order(query: List[Dict]) -> List[Dict]:
    query = deepcopy(query)
    for idx, q in enumerate(query):
        assert 'meta' not in q
        q['meta'] = str(idx)
    return query


def restore_order(results: List[Dict]) -> List[Dict]:
    return sorted(results, key=_sort_key)


def _sort_key(row):
    return int(row['query']['userQuery']['meta'])



# -*- coding: utf-8 -*-
""" Basic command-line interface for AutoExtract. """
import argparse
import json
import sys
from typing import List
import asyncio
import logging

import tqdm

from autoextract.aio import request_parallel, ApiError, create_session
from autoextract.constants import ENV_VARIABLE


logger = logging.getLogger('autoextract')


async def run(urls: List[str], out, n_conn, batch_size, page_type='article',
              api_key=None):
    async with create_session() as session:
        results_iter = request_parallel(urls,
                                        page_type=page_type,
                                        n_conn=n_conn,
                                        batch_size=batch_size,
                                        session=session,
                                        api_key=api_key)
        pbar = tqdm.tqdm(smoothing=0, leave=True, total=len(urls), miniters=1,
                         unit="url")
        try:
            for f in results_iter:
                try:
                    batch_result = await f
                    for res in batch_result:
                        json.dump(res, out, ensure_ascii=False)
                        out.write("\n")
                        out.flush()
                        pbar.update()
                except ApiError as e:
                    logger.error(str(e))
                    raise
        finally:
            pbar.close()


if __name__ == '__main__':
    """ Process urls from input file through AutoExtract """
    p = argparse.ArgumentParser(
        prog='python -m autoextract',
        description="""
        Process input URLs from a file using AutoExtract.
        """,
    )
    p.add_argument("input",
                   type=argparse.FileType("r", encoding='utf8'),
                   help="input file with urls, one url per line")
    p.add_argument("--output", "-o",
                   default=sys.stdout,
                   type=argparse.FileType("w", encoding='utf8'),
                   help=".jsonlines file to store extracted data. "
                        "By default, results are printed to stdout.")
    p.add_argument("--n-conn", type=int, default=20,
                   help="number of connections to the API server "
                        "(default: %(default)s)")
    p.add_argument("--batch-size", type=int, default=2,
                   help="batch size (default: %(default)s)")
    p.add_argument("--page-type", "-t", default="article",
                   help="type of the pages in the input file "
                        "(default: %(default)s)",
                   choices=['article', 'product'])
    p.add_argument("--api-key",
                   help="Scrapinghub AutoExtract API key. "
                        "You can also set %s environment variable instead "
                        "of using this option." % ENV_VARIABLE)
    p.add_argument("--loglevel", "-L", default="INFO",
                   choices=["DEBUG", "INFO", "WARNING", "ERROR"],
                   help="log level")
    args = p.parse_args()
    logging.basicConfig(level=getattr(logging, args.loglevel))

    urls = [u.strip() for u in args.input.readlines() if u.strip()]
    logger.info(f"Loaded {len(urls)} urls from {args.input.name}")
    logger.info(f"Running AutoExtract (connections: {args.n_conn}, "
                f"batch size: {args.batch_size}, page type: {args.page_type})")

    loop = asyncio.get_event_loop()
    coro = run(urls=urls,
               out=args.output,
               n_conn=args.n_conn,
               batch_size=args.batch_size,
               page_type=args.page_type,
               api_key=args.api_key)
    loop.run_until_complete(coro)
    loop.close()



# -*- coding: utf-8 -*-
"""
aiohttp Scrapinghub AutoExtract API client.
"""
import asyncio
from typing import Optional, Dict, Any, List, Iterator
from functools import partial

import aiohttp

from autoextract.batching import record_order, restore_order, build_query
from autoextract.constants import API_ENDPOINT, API_TIMEOUT
from autoextract.apikey import get_apikey
from autoextract.utils import chunks, user_agent
from .retry import autoextract_retry
from .errors import ApiError


AIO_API_TIMEOUT = aiohttp.ClientTimeout(total=API_TIMEOUT + 60,
                                        sock_read=API_TIMEOUT + 30,
                                        sock_connect=10)


def create_session(**kwargs) -> aiohttp.ClientSession:
    """ Create a session with parameters suited for AutoExtract """
    kwargs.setdefault('timeout', AIO_API_TIMEOUT)
    return aiohttp.ClientSession(**kwargs)


async def request_raw(query: List[Dict[str, Any]],
                      api_key: Optional[str] = None,
                      endpoint: str = API_ENDPOINT,
                      *,
                      handle_retries: bool = True,
                      session: Optional[aiohttp.ClientSession] = None
                      ) -> List[Dict]:
    """ Send a request to Scrapinghub AutoExtract API.

    ``query`` is a list of dicts, as described in the API docs
    (see https://doc.scrapinghub.com/autoextract.html).

    ``api_key`` is your AutoExtract API key. If not set, it is
    taken from SCRAPINGHUB_AUTOEXTRACT_KEY environment variable.

    ``session`` is an optional aiohttp.ClientSession object;
    use it to enable HTTP Keep-Alive.

    This function retries http 429 errors and network errors by default;
    this allows to handle server-side throttling properly.
    Use ``handle_retries=False`` if you want to disable this behavior
    (e.g. to implement it yourself).

    See :func:`request_parallel` for a more high-level interface to send
    requests in parallel.
    """
    auth = aiohttp.BasicAuth(get_apikey(api_key))
    headers = {'User-Agent': user_agent(aiohttp)}
    post = _post_func(session)

    async def request():
        async with post(endpoint, json=query, auth=auth, headers=headers) as resp:
            if resp.status >= 400:
                content = await resp.read()
                resp.release()
                raise ApiError(
                    request_info=resp.request_info,
                    history=resp.history,
                    status=resp.status,
                    message=resp.reason,
                    headers=resp.headers,
                    response_content=content
                )
            else:
                return await resp.json()

    if handle_retries:
        request = autoextract_retry(request)

    return await request()


def request_parallel(urls: List[str],
                     page_type: str,
                     api_key: Optional[str] = None,
                     endpoint: str = API_ENDPOINT,
                     *,
                     session: Optional[aiohttp.ClientSession] = None,
                     batch_size=1,
                     n_conn=1,
                     ) -> Iterator[asyncio.Future]:
    """ Send multiple requests to AutoExtract API in parallel.

    ``urls`` is a list of urls to process. All urls are assumed to be
    of type ``page_type`` (e.g. "article").

    ``api_key`` is your AutoExtract API key. If not set, it is
    taken from SCRAPINGHUB_AUTOEXTRACT_KEY environment variable.

    ``n_conn`` is a number of parallel connections to a server.
    ``batch_size`` is an amount of queries sent in a batch in each connection.
    Higher batch_size increase response time, but allows to achieve the same
    throughput with less connections to server.

    ``session`` is an optional aiohttp.ClientSession object;
    use it to enable HTTP Keep-Alive.
    """
    sem = asyncio.Semaphore(n_conn)

    async def _request(batch):
        async with sem:
            return await request_batch(
                urls=batch,
                page_type=page_type,
                api_key=api_key,
                endpoint=endpoint,
                session=session,
            )

    url_batches = chunks(urls, batch_size)
    return asyncio.as_completed([_request(batch) for batch in url_batches])


async def request_batch(urls: List[str],
                        page_type: str,
                        api_key: Optional[str] = None,
                        endpoint: str = API_ENDPOINT,
                        *,
                        handle_retries: bool = True,
                        session: Optional[aiohttp.ClientSession] = None
                        ) -> List[Dict]:
    """ Send a batch request """
    query = record_order(build_query(urls, page_type))
    results = await request_raw(query,
                                api_key=api_key,
                                endpoint=endpoint,
                                session=session,
                                handle_retries=handle_retries)
    return restore_order(results)


def _post_func(session):
    """ Return a function to send a POST request """
    if session is None:
        return partial(aiohttp.request,
                       method='POST',
                       timeout=AIO_API_TIMEOUT)
    else:
        return session.post



# -*- coding: utf-8 -*-
import pytest

from autoextract.apikey import get_apikey, NoApiKey


def test_get_apikey(autoextract_env_variable):
    assert get_apikey('foo') == 'foo'
    assert get_apikey() == autoextract_env_variable


def test_get_apikey_missing():
    with pytest.raises(NoApiKey):
        get_apikey()



# -*- coding: utf-8 -*-
"""
AutoExtract retrying logic.

TODO: add sync support; only autoextract.aio is supported at the moment.
"""
import logging

from aiohttp import client_exceptions
from tenacity import (
    wait_chain,
    wait_fixed,
    wait_random_exponential,
    wait_random,
    stop_after_attempt,
    stop_after_delay,
    retry_if_exception,
    RetryCallState,
    retry,
    before_sleep_log,
    after_log,
)
from tenacity.stop import stop_base, stop_never
from tenacity.wait import wait_base

from .errors import ApiError


logger = logging.getLogger(__name__)


_NETWORK_ERRORS = (
    client_exceptions.ClientResponseError,
    client_exceptions.ClientOSError,
    client_exceptions.ServerConnectionError,
    client_exceptions.ServerDisconnectedError,
    client_exceptions.ServerTimeoutError,
    client_exceptions.ClientPayloadError,
    client_exceptions.ClientConnectorSSLError,
)


def _is_network_error(exc: Exception) -> bool:
    if isinstance(exc, ApiError):
        # ApiError is ClientResponseError, which is in the
        # _NETWORK_ERRORS list, but it should be handled
        # separately.
        return False
    return isinstance(exc, _NETWORK_ERRORS)


def _is_throttling_error(exc: Exception) -> bool:
    return isinstance(exc, ApiError) and exc.status == 429


autoextract_retry_condition = (
    retry_if_exception(_is_throttling_error) |
    retry_if_exception(_is_network_error)
)


class autoextract_wait_strategy(wait_base):
    def __init__(self):
        # throttling
        self.throttling_wait = wait_chain(
            # always wait 20-40s first
            wait_fixed(20) + wait_random(0, 20),

            # wait 20-40s again
            wait_fixed(20) + wait_random(0, 20),

            # wait from 30 to 630s, with full jitter and exponentially
            # increasing max wait time
            wait_fixed(30) + wait_random_exponential(multiplier=1, max=600)
        )

        # connection errors, other client and server failures
        self.network_wait = (
            # wait from 5s to 1m
            wait_fixed(5) + wait_random_exponential(multiplier=1, max=55)
        )

    def __call__(self, retry_state: RetryCallState) -> float:
        exc = retry_state.outcome.exception()
        if _is_throttling_error(exc):
            return self.throttling_wait(retry_state=retry_state)
        elif _is_network_error(exc):
            return self.network_wait(retry_state=retry_state)
        else:
            raise RuntimeError("Invalid retry state exception: %s" % exc)


class autoextract_stop_strategy(stop_base):
    def __init__(self):
        self.stop_on_throttling_error = stop_never
        self.stop_on_network_error = (
            stop_after_attempt(7) |
            stop_after_delay(5 * 60)
        )

    def __call__(self, retry_state: RetryCallState) -> bool:
        exc = retry_state.outcome.exception()
        if _is_throttling_error(exc):
            return self.stop_on_throttling_error(retry_state)
        if _is_network_error(exc):
            return self.stop_on_network_error(retry_state)
        else:
            raise RuntimeError("Invalid retry state exception: %s" % exc)


autoextract_retry = retry(
    wait=autoextract_wait_strategy(),
    retry=autoextract_retry_condition,
    stop=autoextract_stop_strategy(),
    before_sleep=before_sleep_log(logger, logging.DEBUG),
    after=after_log(logger, logging.DEBUG),
)



# -*- coding: utf-8 -*-
from aiohttp import ClientResponseError


class ApiError(ClientResponseError):
    """ Exception which is raised when API returns an error.
    In contrast with ClientResponseError, it allows to inspect response
    content.
    """
    def __init__(self, *args, **kwargs):
        self.response_content = kwargs.pop("response_content")
        super().__init__(*args, **kwargs)

    def __str__(self):
        return f"ApiError: {self.status}, message={self.message}, " \
               f"headers={self.headers}, body={self.response_content}"




#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
from setuptools import setup, find_packages


def get_version():
    about = {}
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, 'autoextract/__version__.py')) as f:
        exec(f.read(), about)
    return about['__version__']


setup(
    name='scrapinghub-autoextract',
    version=get_version(),
    description='Python interface to Scrapinghub Automatic Extraction API',
    long_description=open('README.rst').read() + "\n\n" + open('CHANGES.rst').read(),
    author='Mikhail Korobov',
    author_email='kmike84@gmail.com',
    url='https://github.com/scrapinghub/scrapinghub-autoextract',
    packages=find_packages(exclude=['tests', 'examples']),
    install_requires=[
        'requests',
        'tenacity;python_version>="3.6"',
        'aiohttp >= 3.6.0;python_version>="3.6"',
        'tqdm;python_version>="3.6"',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)



# -*- coding: utf-8 -*-
from .__version__ import __version__


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


def user_agent(library):
    return 'scrapinghub-autoextract/{} {}/{}'.format(
        __version__,
        library.__name__,
        library.__version__)



# -*- coding: utf-8 -*-

# Name of the environment variable with the API key
ENV_VARIABLE = 'SCRAPINGHUB_AUTOEXTRACT_KEY'

# Endpoint
API_ENDPOINT = 'https://autoextract.scrapinghub.com/v1/extract'

# Default timeout that server uses. Client timeouts should be larger than that.
API_TIMEOUT = 600

# Max batch size
API_MAX_BATCH = 100
